@@ -21,6 +21,9 @@ import typing as T
 from contextlib import contextmanager
 from pathlib import Path
 
+if T.TYPE_CHECKING:
+    from ._typing import StringProtocol
+
 """This is (mostly) a standalone module used to write logging
 information about Meson runs. Some output goes to screen,
 some to logging dir and some goes to both."""
@@ -56,7 +59,7 @@ def colorize_console() -> bool:
     sys.stdout.colorize_console = _colorize_console  # type: ignore[attr-defined]
     return _colorize_console
 
-def setup_console():
+def setup_console() -> None:
     # on Windows, a subprocess might call SetConsoleMode() on the console
     # connected to stdout and turn off ANSI escape processing. Call this after
     # running a subprocess to ensure we turn it on again.
@@ -69,7 +72,7 @@ def setup_console():
 log_dir = None               # type: T.Optional[str]
 log_file = None              # type: T.Optional[T.TextIO]
 log_fname = 'meson-log.txt'  # type: str
-log_depth = 0                # type: int
+log_depth = []               # type: T.List[str]
 log_timestamp_start = None   # type: T.Optional[float]
 log_fatal_warnings = False   # type: bool
 log_disable_stdout = False   # type: bool
@@ -97,7 +100,7 @@ def set_verbose() -> None:
 def initialize(logdir: str, fatal_warnings: bool = False) -> None:
     global log_dir, log_file, log_fatal_warnings
     log_dir = logdir
-    log_file = open(os.path.join(logdir, log_fname), 'w', encoding='utf8')
+    log_file = open(os.path.join(logdir, log_fname), 'w', encoding='utf-8')
     log_fatal_warnings = fatal_warnings
 
 def set_timestamp_start(start: float) -> None:
@@ -124,15 +127,38 @@ class AnsiDecorator:
 
     def get_text(self, with_codes: bool) -> str:
         text = self.text
-        if with_codes:
+        if with_codes and self.code:
             text = self.code + self.text + AnsiDecorator.plain_code
         if self.quoted:
-            text = '"{}"'.format(text)
+            text = f'"{text}"'
         return text
 
+    def __len__(self) -> int:
+        return len(self.text)
+
+    def __str__(self) -> str:
+        return self.get_text(colorize_console())
+
+TV_Loggable = T.Union[str, AnsiDecorator, 'StringProtocol']
+TV_LoggableList = T.List[TV_Loggable]
+
+class AnsiText:
+    def __init__(self, *args: TV_LoggableList):
+        self.args = args
+
+    def __len__(self) -> int:
+        return sum(len(x) for x in self.args)
+
+    def __str__(self) -> str:
+        return ''.join(str(x) for x in self.args)
+
+
 def bold(text: str, quoted: bool = False) -> AnsiDecorator:
     return AnsiDecorator(text, "\033[1m", quoted=quoted)
 
+def plain(text: str) -> AnsiDecorator:
+    return AnsiDecorator(text, "")
+
 def red(text: str) -> AnsiDecorator:
     return AnsiDecorator(text, "\033[1;31m")
 
@@ -165,7 +191,7 @@ def normal_cyan(text: str) -> AnsiDecorator:
 
 # This really should be AnsiDecorator or anything that implements
 # __str__(), but that requires protocols from typing_extensions
-def process_markup(args: T.Sequence[T.Union[AnsiDecorator, str]], keep: bool) -> T.List[str]:
+def process_markup(args: T.Sequence[TV_Loggable], keep: bool) -> T.List[str]:
     arr = []  # type: T.List[str]
     if log_timestamp_start is not None:
         arr = ['[{:.3f}]'.format(time.monotonic() - log_timestamp_start)]
@@ -180,7 +206,7 @@ def process_markup(args: T.Sequence[T.Union[AnsiDecorator, str]], keep: bool) ->
             arr.append(str(arg))
     return arr
 
-def force_print(*args: str, **kwargs: T.Any) -> None:
+def force_print(*args: str, nested: str, **kwargs: T.Any) -> None:
     if log_disable_stdout:
         return
     iostr = io.StringIO()
@@ -188,9 +214,13 @@ def force_print(*args: str, **kwargs: T.Any) -> None:
     print(*args, **kwargs)
 
     raw = iostr.getvalue()
-    if log_depth > 0:
-        prepend = '|' * log_depth
-        raw = prepend + raw.replace('\n', '\n' + prepend, raw.count('\n') - 1)
+    if log_depth:
+        prepend = log_depth[-1] + '| ' if nested else ''
+        lines = []
+        for l in raw.split('\n'):
+            l = l.strip()
+            lines.append(prepend + l if l else '')
+        raw = '\n'.join(lines)
 
     # _Something_ is going to get printed.
     try:
@@ -200,7 +230,7 @@ def force_print(*args: str, **kwargs: T.Any) -> None:
         print(cleaned, end='')
 
 # We really want a heterogeneous dict for this, but that's in typing_extensions
-def debug(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
+def debug(*args: TV_Loggable, **kwargs: T.Any) -> None:
     arr = process_markup(args, False)
     if log_file is not None:
         print(*arr, file=log_file, **kwargs)
@@ -209,14 +239,23 @@ def debug(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
 def _debug_log_cmd(cmd: str, args: T.List[str]) -> None:
     if not _in_ci:
         return
-    args = ['"{}"'.format(x) for x in args]  # Quote all args, just in case
+    args = [f'"{x}"' for x in args]  # Quote all args, just in case
     debug('!meson_ci!/{} {}'.format(cmd, ' '.join(args)))
 
 def cmd_ci_include(file: str) -> None:
     _debug_log_cmd('ci_include', [file])
 
-def log(*args: T.Union[str, AnsiDecorator], is_error: bool = False,
+
+def log(*args: TV_Loggable, is_error: bool = False,
+        once: bool = False, **kwargs: T.Any) -> None:
+    if once:
+        return log_once(*args, is_error=is_error, **kwargs)
+    return _log(*args, is_error=is_error, **kwargs)
+
+
+def _log(*args: TV_Loggable, is_error: bool = False,
         **kwargs: T.Any) -> None:
+    nested = kwargs.pop('nested', True)
     arr = process_markup(args, False)
     if log_file is not None:
         print(*arr, file=log_file, **kwargs)
@@ -224,20 +263,26 @@ def log(*args: T.Union[str, AnsiDecorator], is_error: bool = False,
     if colorize_console():
         arr = process_markup(args, True)
     if not log_errors_only or is_error:
-        force_print(*arr, **kwargs)
+        force_print(*arr, nested=nested, **kwargs)
 
-def log_once(*args: T.Union[str, AnsiDecorator], is_error: bool = False,
+def log_once(*args: TV_Loggable, is_error: bool = False,
              **kwargs: T.Any) -> None:
     """Log variant that only prints a given message one time per meson invocation.
 
     This considers ansi decorated values by the values they wrap without
     regard for the AnsiDecorator itself.
     """
-    t = tuple(a.text if isinstance(a, AnsiDecorator) else a for a in args)
+    def to_str(x: TV_Loggable) -> str:
+        if isinstance(x, str):
+            return x
+        if isinstance(x, AnsiDecorator):
+            return x.text
+        return str(x)
+    t = tuple(to_str(a) for a in args)
     if t in _logged_once:
         return
     _logged_once.add(t)
-    log(*args, is_error=is_error, **kwargs)
+    _log(*args, is_error=is_error, **kwargs)
 
 # This isn't strictly correct. What we really want here is something like:
 # class StringProtocol(typing_extensions.Protocol):
@@ -247,16 +292,16 @@ def log_once(*args: T.Union[str, AnsiDecorator], is_error: bool = False,
 # This would more accurately embody what this function can handle, but we
 # don't have that yet, so instead we'll do some casting to work around it
 def get_error_location_string(fname: str, lineno: str) -> str:
-    return '{}:{}:'.format(fname, lineno)
+    return f'{fname}:{lineno}:'
 
-def _log_error(severity: str, *rargs: T.Union[str, AnsiDecorator],
+def _log_error(severity: str, *rargs: TV_Loggable,
                once: bool = False, fatal: bool = True, **kwargs: T.Any) -> None:
     from .mesonlib import MesonException, relpath
 
     # The typing requirements here are non-obvious. Lists are invariant,
     # therefore T.List[A] and T.List[T.Union[A, B]] are not able to be joined
     if severity == 'notice':
-        label = [bold('NOTICE:')]  # type: T.List[T.Union[str, AnsiDecorator]]
+        label = [bold('NOTICE:')]  # type: TV_LoggableList
     elif severity == 'warning':
         label = [yellow('WARNING:')]
     elif severity == 'error':
@@ -274,13 +319,10 @@ def _log_error(severity: str, *rargs: T.Union[str, AnsiDecorator],
         location_str = get_error_location_string(location_file, location.lineno)
         # Unions are frankly awful, and we have to T.cast here to get mypy
         # to understand that the list concatenation is safe
-        location_list = T.cast(T.List[T.Union[str, AnsiDecorator]], [location_str])
+        location_list = T.cast(TV_LoggableList, [location_str])
         args = location_list + args
 
-    if once:
-        log_once(*args, **kwargs)
-    else:
-        log(*args, **kwargs)
+    log(*args, once=once, **kwargs)
 
     global log_warnings_counter
     log_warnings_counter += 1
@@ -288,16 +330,16 @@ def _log_error(severity: str, *rargs: T.Union[str, AnsiDecorator],
     if log_fatal_warnings and fatal:
         raise MesonException("Fatal warnings enabled, aborting")
 
-def error(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
+def error(*args: TV_Loggable, **kwargs: T.Any) -> None:
     return _log_error('error', *args, **kwargs, is_error=True)
 
-def warning(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
+def warning(*args: TV_Loggable, **kwargs: T.Any) -> None:
     return _log_error('warning', *args, **kwargs, is_error=True)
 
-def deprecation(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
+def deprecation(*args: TV_Loggable, **kwargs: T.Any) -> None:
     return _log_error('deprecation', *args, **kwargs, is_error=True)
 
-def notice(*args: T.Union[str, AnsiDecorator], **kwargs: T.Any) -> None:
+def notice(*args: TV_Loggable, **kwargs: T.Any) -> None:
     return _log_error('notice', *args, **kwargs, is_error=False)
 
 def get_relative_path(target: Path, current: Path) -> Path:
@@ -324,7 +366,7 @@ def exception(e: Exception, prefix: T.Optional[AnsiDecorator] = None) -> None:
         # Mypy doesn't follow hasattr, and it's pretty easy to visually inspect
         # that this is correct, so we'll just ignore it.
         path = get_relative_path(Path(e.file), Path(os.getcwd()))  # type: ignore
-        args.append('{}:{}:{}:'.format(path, e.lineno, e.colno))  # type: ignore
+        args.append(f'{path}:{e.lineno}:{e.colno}:')  # type: ignore
     if prefix:
         args.append(prefix)
     args.append(str(e))
@@ -344,10 +386,10 @@ def format_list(input_list: T.List[str]) -> str:
         return ''
 
 @contextmanager
-def nested() -> T.Generator[None, None, None]:
+def nested(name: str = '') -> T.Generator[None, None, None]:
     global log_depth
-    log_depth += 1
+    log_depth.append(name)
     try:
         yield
     finally:
-        log_depth -= 1
+        log_depth.pop()
